

/****************************************************************************************
                               DEMAND INTELLIGENCE

NAME:          di_dm_trans_load.sas

DESCRIPTION:   to load data from a sas data set to a transactional data source

APPLICATIONS:  All

MODULE:        DM

INPUTS:       ALL DI_TRANS parameters should have been initialized;
              upload_table: the table that is uploaded i.e the sas data set
                           that is uploaded from the workspace to the transactional database
              upload_lib : this is the library of the upload_table
              update_table: the table that is updated as part of the sql query
              sql: The update query 
              lock: If the update query updates a large number of records
                    set this to 1,else set this to 0
                    setting it to 1 would update the records after getting
                    an exclusive lock on the tables.
                    
              NOTE: the upload does not support any function calls.
                    It currently supports function call datetime()
                    which is current timestamp in DB2 and now() in MYSQL
                    
                    
               
               
OUTPUTS:       NONE

GLOBAL VAR:    NA

USAGE:        

THREADED:      NA

DOCUMENT:      NA

COMMENTS:  An issue with mysql was identified last week. The problem is with the usage of temporary tables. Temporary tables cannot
Be referenced more than once in any query meaning..a temporary table T1 cannot be in a from clause more than once in the
Sql.
Assume T1 is a temp table.
Eg: Select a.c1 from T1 a, T1 a where a.b1 = a.b1.  (T1 is referenced twice in the query) not valid in mysql
Eg2. update TRANS set trans.a1 = (select a from T1 where a.c1 = trans.c1), set trans.a2 = (Select c from T1 where a.c2 = trans.c2);
       Here we have each of the set clauses have a subselect referencing a temporary table. Mysql does not allow that.

So what we have implemented as a fix for this is to create multiple temporary tables. Copies of the same. And accordingly parsing the
Query and replacing the occurrences of T1 with the temp tables created. PLEASE ENSURE FROM YOUR END THAT WHEN USING
TEMPORARY TABLES in the FROM CLAUSES give ALIASES.
Eg:
. update TRANS set trans.a1 = (select a from T1 where T1.c1 = trans.c1), set trans.a2 = (Select c from T1 where T2.c2 = trans.c2);

SHOULD BE

update TRANS set trans.a1 = (select a from T1 TAB1 where TAB1.c1 = trans.c1), set trans.a2 = (Select c from T1 TAB2  where TAB2.c2 = trans.c2);



AUTHOR:        suyama Yamala,Sumanth

SUPPORT: thburk Burkhardt, Thomas

HISTORY:       started creating macro on Dec05/2005

*****************************************************************************************/

%macro di_dm_trans_update_chunk(upload_lib=,upload_table=,update_table=,sql=,lock=0,debug=0) /*storesecure*/;


/* required params */

%di_util_assert_sym(sym=update_table);

/*
the following data step is to strip out any libnames that might
have been included in the path
*/
data _null_;
str = "&upload_table";
str1=scan(str,2,".");
if str1="" then str1=str;
call symputx("upload_table",str1);
run;

%put upload_table is &upload_table;


%local  to_trans_table l_syscc;

%if "&sql" eq "" %then %do;
   %di_util_abort(macro=di_dm_trans_update,msg='sql cannot be empty');
%end;
%else %do;
   %if (%upcase(&DI_TRANS_VENDOR) eq ORACLE) %then %do;
      %let jobid=&SYSJOBID;
      %local hname l_in_stored_process;

      %let l_in_stored_process = 0;

      %if %symexist(in_stored_process) %then %do;
         %let l_in_stored_process = &in_stored_process;
      %end;
      
      %di_util_get_hostname(name=hname);
         %let hname=%sysfunc(translate(&hname,_,-));
      %if not %symexist(process_sk) %then %do;
          %let process_sk = ;
      %end;
      %if "&l_in_stored_process" eq "1" and "&process_sk" ne "" %then %do;
        %let to_trans_table = tmp_&hname._&process_sk;
      %end;
      %else %do;
       %let to_trans_table = tmp_&hname._&jobid;
      %end;
      
      %put temporary table name is &to_trans_table;
    %end;
    %else %do;
      %let to_trans_table = tmp_upload;
    %end;
    
%end;

%local db_conn upload_lib table_ddl date_columns sastype_date_str;

%di_dm_trans_get_ddl( lib = &upload_lib, table_name=&upload_table, _output_ddl=table_ddl,_date_columns=date_columns, _dbsastype_date=sastype_date_str);
%let date_columns = &date_columns;
 

%if (%upcase(&DI_TRANS_VENDOR) eq DB2 or %upcase(&DI_TRANS_VENDOR) eq ORACLE) %then %do;
   
   %local total_count chunk_size num_iter processed_count i deadlock_loop is_deadlock;
   %let processed_count = 1;
   %let total_count = 0;
   %let deadlock_loop = 0;
   %let is_deadlock = 0;
   
   %if "&upload_table" ne "" %then %do;
      %let total_count = %di_util_nobs(&upload_lib..&upload_table);     
      proc sql noprint;                                                                                                                               
         select spec_value into :chunk_size                                                                                               
         FROM di_data.global_setting                                                                                                 
         where upcase(spec_nm) = 'TRANSACTION_SIZE';
      quit;       
      %if "&chunk_size" eq "" %then %do;
         %let chunk_size = 8192;
      %end;
   
    
      %if &total_count < &chunk_size %then %do;
         %let num_iter = 1;
      %end;
      %else %do;
         %if ( %sysfunc(mod(&total_count,&chunk_size)) eq 0) %then %do;
         %let num_iter = %eval(&total_count/&chunk_size);
         %end;
         %else %do;
            %let num_iter = %eval(&total_count/&chunk_size + 1);
         %end;
      %end;
      %put the value of total_count is &total_count and chunk_size is &chunk_size and num_iter is &num_iter;
      
      %if &total_count > 0 %then %do; /* only if there are any records in upload_table would it update any records */
      %do i = 1 %to &num_iter;
         %if (&i eq &num_iter) %then %do;
           data work.temporary_chunk;
            do PickIt = &processed_count to TotObs by 1;
               set &upload_lib..&upload_table point = PickIt nobs=TotObs;
               output;
            end;
            stop;
           run;
         /* jump to processed_count to end */
            %end;
            %else %do;
               data work.temporary_chunk;
               do PickIt = &processed_count to %eval(&processed_count + &chunk_size -1 ) by 1;
               set &upload_lib..&upload_table point = PickIt;
               output;
               end;
               stop;
               run;
            %end;
            %put the value of processed_count is &processed_count;
            %put the value of chunk_size is &chunk_size;
            %let processed_count = %eval(&processed_count + &chunk_size);
            %put the value of processed_count is &processed_count;
            
            %if (%upcase(&DI_TRANS_VENDOR) eq DB2 ) %then %do;
               %let db_conn = %str(user=&DI_TRANS_USER pwd="&DI_TRANS_PASSWORD"  db=&DI_TRANS_DATABASE);
               libname temp db2 &db_conn schema=SESSION connection=global;
               
               proc sql;
                connect to &DI_TRANS_VENDOR (&db_conn connection=global);
                execute (declare global temporary table &to_trans_table
                        (&table_ddl) 
                           on commit PRESERVE rows not logged) by db2;
               quit;             
            
               proc sql;
                  insert into temp.&to_trans_table select * from work.temporary_chunk;
               quit;  
            
               %if &i eq 1 %then %do;
                  %di_util_searchreplace(initial_string=%bquote(&sql),search=%bquote(&upload_lib..&upload_table),replace=SESSION.&to_trans_table,final_string=sql);
                  %di_util_searchreplace(initial_string=%bquote(&sql),search=datetime(),replace=current timestamp,final_string=sql);      
                  %di_util_searchreplace(initial_string=%bquote(&sql),search=di_trans,replace=&DI_TRANS_SCHEMA,final_string=sql);      
                  %di_util_searchreplace(initial_string=%bquote(&sql),search=DI_TRANS,replace=&DI_TRANS_SCHEMA,final_string=sql);      
                  
               %end;   
               %let l_syscc = &syscc;
               %put the value of l_syscc is &l_syscc;
               proc sql;
                  connect to DB2 as db2_conn (&db_conn connection=global autocommit=no);
                  %if "&lock" eq "1" %then %do;
                     execute ( lock table &DI_TRANS_SCHEMA..&update_table in exclusive mode )by db2_conn;
                  %end;
               
                  execute ( set isolation level cursor stability ) by db2_conn;
                  execute ( &sql ) by db2_conn;
                  %if &sysdbrc eq -911 %then %do; 
                     %let is_deadlock = 1;
                     execute (rollback) by db2_conn; 
                  %end;
                  %else %do;
                     execute( commit )by db2_conn;
                     execute ( drop table SESSION.&to_trans_table) by db2_conn; 
                     execute (commit) by db2_conn; 
                  %end;
                  disconnect from db2_conn;
               quit;
               /* fail safe code ..executes only once and sets lock to 1 once
                   a deadlock is detected
               */
               %if &is_deadlock eq 1 and &deadlock_loop eq 0 %then %do;
                  %let deadlock_loop = 1; /* this is to prevent a long loop if 
                                         using an exclusive lock also results in a deadlock;*/
                  %let syscc = &l_syscc;
                  /* setting the value of syscc to a value it was before the deadlock
                  was identified
                  */
                  %let lock =1; 
                  proc sql;
                     connect to DB2 as db2_conn (&db_conn connection=global autocommit=no);
                     execute ( lock table &DI_TRANS_SCHEMA..&update_table in exclusive mode )by db2_conn;
                     execute ( set isolation level cursor stability ) by db2_conn;
                     execute ( &sql ) by db2_conn;
                     execute( commit )by db2_conn;
                     execute ( drop table SESSION.&to_trans_table) by db2_conn; 
                     execute (commit) by db2_conn; 
                     disconnect from db2_conn;
                  quit;
                                        
               %end;
            %end ; /* end of DB2 */
            %else %if (%upcase(&DI_TRANS_VENDOR) eq ORACLE) %then %do;
               %let db_conn = %str(user=&DI_TRANS_USER pw="&DI_TRANS_PASSWORD"  path=&DI_TRANS_DATABASE);
               libname temp oracle &db_conn connection=global insertbuff=32767;
               
               %put to_trans_table is &to_trans_table;
               
               %if %sysfunc (exist(temp.&to_trans_table)) %then %do;
                  proc sql;
                     drop table temp.&to_trans_table;
                  quit;
               %end;
               
               proc sql;
                connect to &DI_TRANS_VENDOR (&db_conn connection=global);
                execute (create global temporary table &to_trans_table
                        (&table_ddl) 
                           on commit PRESERVE rows ) by oracle;
               quit;
            
               %if "&sastype_date_str" ne "" %then %do;                 
                  proc sql;
                     insert into temp.&to_trans_table (DBSASTYPE=(&sastype_date_str) )
                     select * from work.temporary_chunk;
                  quit; 
               %end;
               %else %do;
                  proc sql;
                     insert into temp.&to_trans_table select * from work.temporary_chunk;
                  quit;  
               %end;
            
            
               %if &i eq 1 %then %do;
               %di_util_searchreplace(initial_string=%bquote(&sql),search=%bquote(&upload_lib..&upload_table),replace=&to_trans_table,final_string=sql);
               %di_util_searchreplace(initial_string=%bquote(&sql),search=datetime(),replace=sysdate,final_string=sql);      
               %di_util_searchreplace(initial_string=%bquote(&sql),search=di_trans,replace=&DI_TRANS_SCHEMA,final_string=sql);      
	       %di_util_searchreplace(initial_string=%bquote(&sql),search=DI_TRANS,replace=&DI_TRANS_SCHEMA,final_string=sql);      
               %end;   
            
               proc sql;
                  connect to ORACLE as ora_conn (&db_conn connection=global );
                        
                  execute ( &sql ) by ora_conn;
                  execute( commit )by ora_conn;
                  disconnect from ora_conn;
               quit;
            %end;
            
            libname temp clear;  
               
            %if (%upcase(&DI_TRANS_VENDOR) eq ORACLE) %then %do;
                proc sql;
                 connect to ORACLE as ora_conn (&db_conn connection=global );
                 execute ( drop table &to_trans_table ) by ora_conn;
                 execute( commit )by ora_conn;
                 disconnect from ora_conn;
                quit;
            %end;
            
            %if &syscc > 4 %then %do;   
               %goto EXIT;
            %end;

      %end;/* to be in the end do i=1 to num_iter*/
      %end; /* end of total_count > 0 */
      
      
   %end; 
   %else %do; /* no upload table */ 
      %if (%upcase(&DI_TRANS_VENDOR) eq DB2 )%then %do; 
         %di_util_searchreplace(initial_string=%bquote(&sql),search=datetime(),replace=current timestamp,final_string=sql);      
         %di_util_searchreplace(initial_string=%bquote(&sql),search=di_trans,replace=&DI_TRANS_SCHEMA,final_string=sql);      
         %di_util_searchreplace(initial_string=%bquote(&sql),search=DI_TRANS,replace=&DI_TRANS_SCHEMA,final_string=sql);      
         proc sql;
            connect to DB2 as db2_conn (&db_conn connection=global autocommit=no);
            %if "&lock" eq "1" %then %do;
            execute ( lock table &DI_TRANS_SCHEMA..&update_table in exclusive mode )by db2_conn;
            %end;
            execute ( set isolation level cursor stability ) by db2_conn;
            execute ( &sql ) by db2_conn;
            execute( commit )by db2_conn;
            disconnect from db2_conn;
         quit;
      %end;
      %else %if (%upcase(&DI_TRANS_VENDOR) eq ORACLE) %then %do;
         %di_util_searchreplace(initial_string=%bquote(&sql),search=datetime(),replace=sysdate,final_string=sql);      
         %di_util_searchreplace(initial_string=%bquote(&sql),search=di_trans,replace=&DI_TRANS_SCHEMA,final_string=sql);      
         %di_util_searchreplace(initial_string=%bquote(&sql),search=DI_TRANS,replace=&DI_TRANS_SCHEMA,final_string=sql);      
         proc sql;
            connect to ORACLE as ora_conn (&db_conn connection=global );
            execute ( &sql ) by ora_conn;
            execute( commit )by ora_conn;
            disconnect from ora_conn;
         quit;
      %end;
   %end;   
%end; 
%else %do;/* a lock is applied because when run in parallel it results in lock issues if SAS is trans vendor */
   %di_util_searchreplace(initial_string=%bquote(&sql),search=%nrstr(&DI_TRANS_SCHEMA.),replace=di_trans,final_string=sql);      
   %di_util_lock( di_trans.&update_table );
   proc sql;
     &sql;
   quit;
   %di_util_unlock( di_trans.&update_table );
%end;

%EXIT:

%put the value of di_dm_trans_update_chunk syscc is &SYSCC;
%mend di_dm_trans_update_chunk;
